#include "RawCamBehavior.h"
#include "Wireless/Wireless.h"
#include "Events/EventRouter.h"
#include "Vision/RawCameraGenerator.h"
#include "Vision/JPEGGenerator.h"
#include "Events/FilterBankEvent.h"
#include "Behaviors/Controller.h"
#include "Shared/ProjectInterface.h"

RawCamBehavior::RawCamBehavior()
	: BehaviorBase(), visRaw(NULL), packet(NULL), cur(NULL), avail(0), max_buf(0)
{}

void
RawCamBehavior::DoStart() {
	BehaviorBase::DoStart();
	
	std::vector<std::string> args;
	if(config->vision.rawcam_transport==0) {
		max_buf=UDP_WIRELESS_BUFFER_SIZE;
		visRaw=wireless->socket(SocketNS::SOCK_DGRAM, 1024, max_buf);
		args.push_back("udp");
	} else if(config->vision.rawcam_transport==1) {
		max_buf=TCP_WIRELESS_BUFFER_SIZE;
		visRaw=wireless->socket(SocketNS::SOCK_STREAM, 1024, max_buf);
		wireless->setDaemon(visRaw,true);
		args.push_back("tcp");
	} else {
		serr->printf("ERROR: Invalid Config::vision.rawcam_transport: %d\n",config->vision.rawcam_transport);
		return;
	}
	wireless->listen(visRaw,config->vision.rawcam_port);
	
	Controller::loadGUI("org.tekkotsu.mon.VisionGUI","RawVisionGUI",config->vision.rawcam_port,args);

	erouter->addListener(this,EventBase::visRawCameraEGID,ProjectInterface::visRawCameraSID);
	erouter->addListener(this,EventBase::visJPEGEGID,ProjectInterface::visColorJPEGSID);
	erouter->addListener(this,EventBase::visJPEGEGID,ProjectInterface::visGrayscaleJPEGSID);
}

void
RawCamBehavior::DoStop() {
	erouter->removeListener(this);
	Controller::closeGUI("RawVisionGUI");

	// this could be considered a bug in our wireless - if we don't setDaemon(...,false)
	// it will try to listen again even though we explicitly closed the server socket...
	wireless->setDaemon(visRaw,false);
	wireless->close(visRaw->sock);
	BehaviorBase::DoStop();
}

void
RawCamBehavior::processEvent(const EventBase& e) {
	if(!wireless->isConnected(visRaw->sock))
		return;
	const FilterBankEvent* fbke=dynamic_cast<const FilterBankEvent*>(&e);
	ASSERTRET(fbke!=NULL,"unexpected event");
	/* // turning these off enables individual channel compression
		if(config->vision.rawcam_compression==Config::vision_config::COMPRESS_NONE && e.getGeneratorID()!=EventBase::visRawCameraEGID)
		return;
		if(config->vision.rawcam_compression==Config::vision_config::COMPRESS_JPEG && e.getGeneratorID()!=EventBase::visJPEGEGID)
		return; */
	if(config->vision.rawcam_encoding==Config::vision_config::ENCODE_COLOR) {
		bool succ=writeColor(*fbke);
		ASSERTRET(succ,"serialization failed");
	} else if(config->vision.rawcam_encoding==Config::vision_config::ENCODE_SINGLE_CHANNEL) {
		bool succ=writeSingleChannel(*fbke);
		ASSERTRET(succ,"serialization failed");
	}
}

bool
RawCamBehavior::openPacket(FilterBankGenerator& fbkgen, unsigned int time, unsigned int layer) {
	if(packet!=NULL)
		return false;

	avail=max_buf-1; //not sure why -1, but Alok had it, so i will too
	ASSERT(cur==NULL,"cur non-NULL");
	cur=NULL;
	char * buf=packet=(char*)visRaw->getWriteBuffer(avail);
	ASSERTRETVAL(packet!=NULL,"could not get buffer",false);
	
	unsigned int used;
	used=LoadSave::encode("TekkotsuImage",buf,avail);
	ASSERTRETVAL(used!=0,"ran out of space",false);
	avail-=used; buf+=used;
	used=LoadSave::encode(config->vision.rawcam_encoding,buf,avail);
	ASSERTRETVAL(used!=0,"ran out of space",false);
	avail-=used; buf+=used;
	used=LoadSave::encode(config->vision.rawcam_compression,buf,avail);
	ASSERTRETVAL(used!=0,"ran out of space",false);
	avail-=used; buf+=used;

	used=LoadSave::encode(fbkgen.getWidth(layer),buf,avail);
	ASSERTRETVAL(used!=0,"ran out of space",false);
	avail-=used; buf+=used;
	used=LoadSave::encode(fbkgen.getHeight(layer),buf,avail);
	ASSERTRETVAL(used!=0,"ran out of space",false);
	avail-=used; buf+=used;
	used=LoadSave::encode(time,buf,avail);
	ASSERTRETVAL(used!=0,"ran out of space",false);
	avail-=used; buf+=used;
	used=LoadSave::encode(fbkgen.getFrameNumber(),buf,avail);
	ASSERTRETVAL(used!=0,"ran out of space",false);
	avail-=used; buf+=used;

	cur=buf;
	return true;
}

bool
RawCamBehavior::writeColor(const FilterBankEvent& e) {
	FilterBankGenerator& fbkgen=*e.getSource();

	unsigned int y_layer=fbkgen.getNumLayers()-1-config->vision.rawcam_y_skip;
	unsigned int uv_layer=fbkgen.getNumLayers()-1-config->vision.rawcam_uv_skip;

	if(config->vision.rawcam_channel==-1) {
		if(NULL!=dynamic_cast<const RawCameraGenerator*>(&fbkgen) && config->vision.rawcam_compression==Config::vision_config::COMPRESS_NONE
			 || NULL!=dynamic_cast<const JPEGGenerator*>(&fbkgen) && config->vision.rawcam_compression==Config::vision_config::COMPRESS_JPEG) {
			if(const JPEGGenerator* jgen=dynamic_cast<const JPEGGenerator*>(&fbkgen))
				if(jgen->getCurrentSourceFormat()==JPEGGenerator::SRC_COLOR)
					return true;
			unsigned int used=0;
			openPacket(fbkgen,e.getTimeStamp(),uv_layer);
			ASSERTRETVAL(cur!=NULL,"header failed",false);
			
			used=LoadSave::encode("FbkImage",cur,avail);
			ASSERTRETVAL(used!=0,"save blank image failed",false);
			avail-=used; cur+=used;
			used=LoadSave::encode(0,cur,avail);
			ASSERTRETVAL(used!=0,"save blank image failed",false);
			avail-=used; cur+=used;
			used=LoadSave::encode(0,cur,avail);
			ASSERTRETVAL(used!=0,"save blank image failed",false);
			avail-=used; cur+=used;
			used=LoadSave::encode(-1,cur,avail);
			ASSERTRETVAL(used!=0,"save blank image failed",false);
			avail-=used; cur+=used;
			used=LoadSave::encode(RawCameraGenerator::CHAN_Y,cur,avail);
			ASSERTRETVAL(used!=0,"save blank image failed",false);
			avail-=used; cur+=used;
			used=LoadSave::encode("blank",cur,avail);
			ASSERTRETVAL(used!=0,"save blank image failed",false);
			avail-=used; cur+=used;

			fbkgen.selectSaveImage(uv_layer,RawCameraGenerator::CHAN_U);
			used=fbkgen.SaveBuffer(cur,avail);
			ASSERTRETVAL(used!=0,"save image failed",false);
			avail-=used; cur+=used;
			
			fbkgen.selectSaveImage(uv_layer,RawCameraGenerator::CHAN_V);
			used=fbkgen.SaveBuffer(cur,avail);
			ASSERTRETVAL(used!=0,"save image failed",false);
			avail-=used; cur+=used;

			closePacket();
		}
		return true;
	}

	unsigned int big_layer=y_layer;
	unsigned int small_layer=uv_layer;
	if(y_layer<uv_layer) { 
		big_layer=uv_layer;
		small_layer=y_layer;
	}
	if(dynamic_cast<const RawCameraGenerator*>(&fbkgen)) {
		unsigned int used=0;
		bool opened=false;

		if(config->vision.rawcam_compression==Config::vision_config::COMPRESS_NONE || big_layer-small_layer>=2 && big_layer==uv_layer) {
			opened=openPacket(fbkgen,e.getTimeStamp(),big_layer);
			ASSERTRETVAL(cur!=NULL,"header failed",false);
			
			fbkgen.selectSaveImage(y_layer,RawCameraGenerator::CHAN_Y);
			used=fbkgen.SaveBuffer(cur,avail);
			ASSERTRETVAL(used!=0,"save image failed",false);
			avail-=used; cur+=used;
		}

		if(config->vision.rawcam_compression==Config::vision_config::COMPRESS_NONE || big_layer-small_layer>=2 && big_layer==y_layer) {
			opened=openPacket(fbkgen,e.getTimeStamp(),big_layer);
			ASSERTRETVAL(cur!=NULL,"header failed",false);
			
			fbkgen.selectSaveImage(uv_layer,RawCameraGenerator::CHAN_U);
			used=fbkgen.SaveBuffer(cur,avail);
			ASSERTRETVAL(used!=0,"save image failed",false);
			avail-=used; cur+=used;
			
			fbkgen.selectSaveImage(uv_layer,RawCameraGenerator::CHAN_V);
			used=fbkgen.SaveBuffer(cur,avail);
			ASSERTRETVAL(used!=0,"save image failed",false);
			avail-=used; cur+=used;
		}

		if(config->vision.rawcam_compression==Config::vision_config::COMPRESS_NONE || !opened)
			closePacket();
	} else if(const JPEGGenerator* jgen=dynamic_cast<const JPEGGenerator*>(&fbkgen)) {
		if(config->vision.rawcam_compression!=Config::vision_config::COMPRESS_JPEG)
			return true;
		if(jgen->getCurrentSourceFormat()==JPEGGenerator::SRC_COLOR && big_layer-small_layer<2) {
			unsigned int used=0;
			openPacket(fbkgen,e.getTimeStamp(),big_layer);
			ASSERTRETVAL(cur!=NULL,"header failed",false);
			
			fbkgen.selectSaveImage(big_layer,0);
			used=fbkgen.SaveBuffer(cur,avail);
			ASSERTRETVAL(used!=0,"save image failed",false);
			avail-=used; cur+=used;
			
			closePacket();
		} else if(jgen->getCurrentSourceFormat()==JPEGGenerator::SRC_GRAYSCALE && big_layer-small_layer>=2) {
			unsigned int used=0;
			bool opened=openPacket(fbkgen,e.getTimeStamp(),big_layer);
			ASSERTRETVAL(cur!=NULL,"header failed",false);
			
			if(big_layer==y_layer) {
				fbkgen.selectSaveImage(y_layer,RawCameraGenerator::CHAN_Y);
				used=fbkgen.SaveBuffer(cur,avail);
				ASSERTRETVAL(used!=0,"save image failed",false);
				avail-=used; cur+=used;
			} else {
				fbkgen.selectSaveImage(uv_layer,RawCameraGenerator::CHAN_U);
				used=fbkgen.SaveBuffer(cur,avail);
				ASSERTRETVAL(used!=0,"save image failed",false);
				avail-=used; cur+=used;
			
				fbkgen.selectSaveImage(uv_layer,RawCameraGenerator::CHAN_V);
				used=fbkgen.SaveBuffer(cur,avail);
				ASSERTRETVAL(used!=0,"save image failed",false);
				avail-=used; cur+=used;
			}
			
			if(!opened)
				closePacket();
		}
	}

	return true;
}

bool
RawCamBehavior::writeSingleChannel(const FilterBankEvent& e) {
	FilterBankGenerator& fbkgen=*e.getSource();
	if( config->vision.rawcam_compression==Config::vision_config::COMPRESS_NONE && dynamic_cast<const RawCameraGenerator*>(&fbkgen)
			|| config->vision.rawcam_compression==Config::vision_config::COMPRESS_JPEG && dynamic_cast<const JPEGGenerator*>(&fbkgen) )
		{
			if(const JPEGGenerator * jgen=dynamic_cast<const JPEGGenerator*>(&fbkgen))
				if(jgen->getCurrentSourceFormat()!=JPEGGenerator::SRC_GRAYSCALE)
					return true;
			unsigned int layer=fbkgen.getNumLayers()-1-config->vision.rawcam_y_skip;
			
			unsigned int used=0;
			openPacket(fbkgen,e.getTimeStamp(),layer);
			ASSERTRETVAL(cur!=NULL,"header failed",false);
			
			fbkgen.selectSaveImage(layer,config->vision.rawcam_channel);
			used=fbkgen.SaveBuffer(cur,avail);
			ASSERTRETVAL(used!=0,"save image failed",false);
			avail-=used; cur+=used;
			
			closePacket();
		}
	return true;
}

//#include "Shared/WorldState.h"

void
RawCamBehavior::closePacket() {
	if(packet==NULL)
		return;
	visRaw->write(cur-packet);
	/*	cout << "used=" << (cur-packet) << "\tavail==" << avail;
	if(state->robotDesign & WorldState::ERS7Mask)
		cout << "\tmax bandwidth=" << (cur-packet)/1024.f*30 << "KB/sec" << endl;
	else
		cout << "\tmax bandwidth=" << (cur-packet)/1024.f*24 << "KB/sec" << endl;
	*/
	packet=cur=NULL;
	avail=0;
}


/*! @file
 * @brief Implements RawCamBehavior, which forwards images from camera over wireless
 * @author ejt (Creator)
 *
 * $Author: ejt $
 * $Name: tekkotsu-2_2 $
 * $Revision: 1.13 $
 * $State: Exp $
 * $Date: 2004/10/07 19:07:05 $
 */

